"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fakeSelection = exports.cmReplayMultiple = exports.cmReplay = exports.selectCmd = void 0;
const state_1 = require("@codemirror/state");
const fake_selection_1 = require("./fake-selection.cjs");
/** Reserved command for specifying file. */
exports.selectCmd = "file:";
/**
 * Replay typing in CodeMirror.
 * @returns Unsubscription function.
 */
function cmReplay({ data, handle, playback, start, view, }) {
    return cmReplayMultiple({
        data: [[0, exports.selectCmd + "default"], ...data],
        handle: (key, docs) => handle(key, docs["default"]),
        playback,
        start,
        views: {
            default: view,
        },
    });
}
exports.cmReplay = cmReplay;
/**
 * Replay typing to several CodeMirror instances in parallel.
 * @returns Unsubscription function.
 */
function cmReplayMultiple({ data, handle, playback, start = 0, views, }) {
    /** Current file being replayed into */
    let file = undefined;
    // validation
    if (!(data.length > 0 &&
        data[0][0] === 0 &&
        typeof data[0][1] === "string" &&
        data[0][1].startsWith(exports.selectCmd))) {
        throw new Error("First command must have time 0 and select the file");
    }
    // we're going to mess with data, clone it
    data = JSON.parse(JSON.stringify(data));
    /* unpackage */
    // decompress times
    const times = data.map((_) => _[0]);
    for (let i = 1; i < times.length; ++i)
        times[i] += times[i - 1];
    // deserialize changesets
    for (const entry of data) {
        if (typeof entry[1] !== "string")
            entry[1][0] = state_1.ChangeSet.fromJSON(entry[1][0]);
    }
    // initialize inverses
    const inverses = {};
    for (const key in views) {
        inverses[key] = [];
    }
    // compute inverses
    {
        const docs = {};
        for (const key in views) {
            docs[key] = views[key].state.doc;
        }
        for (let i = 0; i < data.length; ++i) {
            const action = data[i][1];
            if (typeof action !== "string") {
                inverses[file][i] = action[0].invert(docs[file]);
                docs[file] = action[0].apply(docs[file]);
            }
            else if (action.startsWith(exports.selectCmd)) {
                file = action.slice(exports.selectCmd.length);
            }
        }
    }
    /* main logic */
    let index = 0;
    let lastTime = 0;
    const repaint = () => {
        const t = playback.currentTime;
        const progress = (t - start) * 1000;
        const changes = {};
        for (const key in views) {
            changes[key] = state_1.ChangeSet.empty(views[key].state.doc.length);
        }
        const selections = {};
        // apply / revert changes
        if (lastTime <= t && index < data.length) {
            let i = index;
            for (; i < data.length && times[i] <= progress; ++i) {
                const action = data[i][1];
                if (typeof action === "string") {
                    if (action.startsWith(exports.selectCmd)) {
                        file = action.slice(exports.selectCmd.length);
                    }
                    // handle action
                    const docs = {};
                    for (const key in views) {
                        docs[key] = changes[key].apply(views[key].state.doc);
                    }
                    handle(action, docs);
                }
                else {
                    changes[file] = changes[file].compose(action[0]);
                    // handle selection
                    if (action[1]) {
                        const [anchor, head] = action[1];
                        selections[file] = { anchor, head };
                    }
                }
            }
            index = i;
        }
        else if (t < lastTime && 0 < index) {
            let i = index - 1;
            for (; 0 <= i && progress < times[i]; --i) {
                if (inverses[file][i]) {
                    changes[file] = changes[file].compose(inverses[file][i]);
                }
                else if (data[i][1] === exports.selectCmd + file) {
                    // find file to replay into
                    for (let j = i - 1; 0 <= j; --j) {
                        const action = data[j][1];
                        if (typeof action === "string" && action.startsWith(exports.selectCmd)) {
                            file = action.slice(exports.selectCmd.length);
                            // XXX figure out how to handle more general actions
                            handle(action, {});
                            break;
                        }
                    }
                }
            }
            index = i + 1;
        }
        for (const key in views) {
            const effects = selections[key] ? [fake_selection_1.FakeSelection.of(selections[key])] : undefined;
            views[key].dispatch(views[key].state.update({ changes: changes[key], effects }));
        }
        lastTime = t;
    };
    /* subscribe */
    playback.addEventListener("seeking", repaint);
    playback.addEventListener("timeupdate", repaint);
    return () => {
        playback.removeEventListener("seeking", repaint);
        playback.removeEventListener("timeupdate", repaint);
    };
}
exports.cmReplayMultiple = cmReplayMultiple;
var fake_selection_2 = require("./fake-selection.cjs");
Object.defineProperty(exports, "fakeSelection", { enumerable: true, get: function () { return fake_selection_2.fakeSelection; } });
